#~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*#
# thumbnails.tcl                                                           #
#                                                                          #
# Description:                                                             #
#                                                                          #
#     This Tcl/Tk script defines an inteface for displaying thumbnail      #
#     images of hand-drawn symbols, and for animating the symbols to       #
#     show the order in which the strokes are to be drawn.                 #
#                                                                          #
# Changes:                                                                 #
#                                                                          #
#     09/19/99  arvo    Initial coding.                                    #
#                                                                          #
#--------------------------------------------------------------------------#
# Copyright (C) 1999, James Arvo                                           #
#                                                                          #
# This program is free software; you can redistribute it and/or modify it  #
# under the terms of the GNU General Public License as published by the    #
# Free Software Foundation.  See http://www.fsf.org/copyleft/gpl.html      #
#                                                                          #
# This program is distributed in the hope that it will be useful, but      #
# WITHOUT EXPRESS OR IMPLIED WARRANTY of merchantability or fitness for    #
# any particular purpose.  See the GNU General Public License for more     #
# details.                                                                 #
#                                                                          #
#~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*#

proc Thumb_Begin { {frame "" } } {
  global thumb colors
  set destroy .
  if { $frame == "" } then {
    set frame .thumbnails
    if [catch "toplevel $frame -visual best"] then return
    wm title $frame "Thumbnails of Recognized Symbols"
    set destroy $frame
    }

  set bgc $colors(background)
  set box $frame.box

  frame  $box             -relief flat -borderwidth 0
  frame  $box.header      -relief ridge -borderwidth 3 -background white
  label  $box.header.symb -text "" -relief flat -borderwidth 0 -background white
  label  $box.header.file -text "" -relief flat -borderwidth 0 -background white
  frame  $box.bev         -relief sunken -borderwidth 3 -background $bgc
  frame  $box.bar         -relief flat -borderwidth 0
  button $box.bar.help    -text Help   -relief raised -bd 3 -state disabled
  button $box.bar.resize  -text Resize -relief raised -bd 3 -state disabled
  button $box.bar.close   -text Close  -relief raised -bd 3 -command "destroy $destroy"

  set thumb(width)  200               ;# The default width of the canvas.
  set thumb(height) 300               ;# The default height of the canvas.
  set thumb(ncol)   10                ;# The number of symbols displayed on a row.
  set thumb(pad)    6                 ;# Padding between canvas and beveled frame.
  set thumb(row)    0                 ;# Counter for the current row.
  set thumb(col)    0                 ;# Counter for the current column.
  set thumb(head1)  $box.header.symb  ;# Dynamic readout of symbol name.
  set thumb(head2)  $box.header.file  ;# Dynamic readout of file of definition.
  set thumb(bev)    $box.bev          ;# Bevel around the scrollable canvas.
  set thumb(canv)   $thumb(bev).canv  ;# The scrollable canvas itself.
  set thumb(grid)   $thumb(canv).grid ;# The grid that holds all the thumbnails.
  set thumb(xslide) $box.xslide       ;# The horizontal slider.
  set thumb(yslide) $box.yslide       ;# The vertical slider.

  canvas $thumb(canv)      \
    -width  $thumb(width)  \
    -height $thumb(height) \
    -background $bgc       \
    -relief flat           \
    -borderwidth 0         \
    -confine true          \
    -scrollregion {0 0 100 100}

  bind $thumb(canv) <Configure> Thumb_Reconfigure_Sliders

  scale $thumb(yslide) \
    -relief sunken     \
    -borderwidth 1     \
    -from 0            \
    -to 1.0            \
    -resolution 0.01   \
    -orient vertical   \
    -width 16          \
    -showvalue false   \
    -sliderlength 50   \
    -command Thumb_Y_Slider

  scale $thumb(xslide) \
    -relief sunken     \
    -borderwidth 1     \
    -from 0            \
    -to 1.0            \
    -resolution 0.01   \
    -orient horizontal \
    -width 16          \
    -showvalue false   \
    -sliderlength 50   \
    -command Thumb_X_Slider

  pack  $box.header.symb -side top -pady 4
  pack  $box.header.file -side top -pady 4
  pack  $box.bar.help    -side left
  pack  $box.bar.resize  -side left -padx 2
  pack  $box.bar.close   -side left -fill x -expand true

  pack $box -side top -expand true -fill both -padx 10 -pady 10
  pack $box.bev.canv -side top -expand true -fill both -padx $thumb(pad) -pady $thumb(pad)
  grid $box.header -sticky nsew -columnspan 2 -pady 6
  grid $box.bev $thumb(yslide) -sticky nsew 
  grid $thumb(xslide) -sticky nsew
  grid $box.bar -sticky ew -columnspan 2 -pady 8
  grid rowconfigure    $box 1 -weight 1
  grid columnconfigure $box 0 -weight 1

  frame $thumb(grid) -background $bgc
  $thumb(canv) create window 0 0 -anchor nw -window $thumb(grid)
  }

proc Thumb_Section { dir_name {full_dir_name ""} } {
  global thumb
  incr thumb(row)
  set sec  sec$thumb(row)
  set path $thumb(grid).$sec
  label $path -relief ridge -bd 3 -text $dir_name 
  grid $path -row $thumb(row) -columnspan $thumb(ncol) -padx 2 -pady 14 -ipadx 6 -ipady 4
  if { $full_dir_name != "" } then {
    bind $path <Enter> "$path configure -text $full_dir_name"
    bind $path <Leave> "$path configure -text $dir_name"
    }
  incr thumb(row)
  set thumb(col) 0
  }

proc Thumb_AddImage { res symbol_name file_name } {
  global thumb
  if { $thumb(col) == $thumb(ncol) } then {
    set thumb(col) 0
    incr thumb(row)
    Thumb_End  ;# Update the interface as sections are added, to show progress.
    }
  set img symbol_$thumb(row)_$thumb(col)
  set thumb(symb) $thumb(grid).$img
  canvas $thumb(symb)           \
    -height $res                \
    -width  $res                \
    -background white           \
    -relief raised              \
    -borderwidth 2              \
    -highlightthickness 2       \
    -highlightbackground gray50 
  bind $thumb(symb) <ButtonPress-1> "SendMessage thumb_anim $symbol_name $file_name"
  grid $thumb(symb) -row $thumb(row) -column $thumb(col) -padx 1 -pady 1
  bind $thumb(symb) <Enter> \
    "$thumb(head1) configure -text $symbol_name
     $thumb(head2) configure -text $file_name
     $thumb(symb)  configure -highlightbackground orange -relief sunken"
  bind $thumb(symb) <Leave> \
    "$thumb(head1) configure -text {}
     $thumb(head2) configure -text {}
     $thumb(symb)  configure -highlightbackground gray50 -relief raised"
  incr thumb(col)
  }

proc Thumb_Line { x0 y0 x1 y1 } {
  global thumb
  $thumb(symb) create line $x0 $y0 $x1 $y1 -width 2
  }

proc Thumb_End {} {
  global thumb
  update
  $thumb(canv) configure              \
    -width [winfo width $thumb(grid)] \
    -scrollregion [grid bbox $thumb(grid)]
  Thumb_Reconfigure_Sliders
  }

proc Thumb_Animate { symb file } {
  global thumb
  if [catch "toplevel .anim -visual best"] then {
   .anim.label configure -text $symb
    Thumb_Replay
    return
    }
  wm title .anim "show me"

  canvas .anim.canv    \
    -height 150        \
    -width  150        \
    -background white  \
    -relief sunken     \
    -borderwidth 3

  label  .anim.label     -text "$symb" -relief flat -bd 0 
  frame  .anim.row       -relief flat 
  button .anim.row.play  -text replay -command Thumb_Replay
  button .anim.row.close -text close  -command {destroy .anim}
  pack   .anim.row.play  -side left -fill x -expand true
  pack   .anim.row.close -side left -fill x -expand true
  pack   .anim.label -side top -fill x -expand true -padx 6 -pady 6
  pack   .anim.canv  -side top                      -padx 6
  pack   .anim.row   -side top -fill x -expand true -padx 6 -pady 6
  Thumb_Replay
  }

proc Thumb_Replay {} {
 .anim.canv delete ALL
  SendMessage thumb_anim
  }

proc Thumb_Anim_Line { wait x0 y0 x1 y1 } {
  after $wait ".anim.canv create line $x0 $y0 $x1 $y1 -width 3 -tag ALL; update"
  }



#
# The following procedures are used internally -- they are not to
# be called by the application.
#
proc Thumb_Reconfigure_Sliders {} {
  global thumb
  set x_needed [winfo width  $thumb(grid)]
  set y_needed [winfo height $thumb(grid)]
  set x_size   [winfo width  $thumb(canv)]
  set y_size   [winfo height $thumb(canv)]
  set x_len    [winfo width  $thumb(bev)]
  set y_len    [winfo height $thumb(bev)]
  set x_limit  1.0
  set y_limit  1.0
  set x_frac   1.0
  set y_frac   1.0
  if { $x_size < $x_needed } then {
    set x_limit [expr ($x_needed - $x_size) / (0.975 * $x_needed)]
    set x_frac  [expr $x_size / (1.0 * $x_needed)]
    }
  if { $y_size < $y_needed } then {
    set y_limit [expr ($y_needed - $y_size) / (0.975 * $y_needed)]
    set y_frac  [expr $y_size / (1.0 * $y_needed)]
    }
  $thumb(xslide) configure -sliderlength [expr $x_frac * $x_len - 6] -to $x_limit
  $thumb(yslide) configure -sliderlength [expr $y_frac * $y_len - 6] -to $y_limit
  }

proc Thumb_Y_Slider { y } {
  global thumb
  $thumb(canv) yview moveto $y 
  }

proc Thumb_X_Slider { x } {
  global thumb
  $thumb(canv) xview moveto $x
  }

