/***************************************************************************
* File.h                                                                   *
*                                                                          *
* Tools for traversing directory structures.                               *
*                                                                          *
*   HISTORY                                                                *
*       09/22/99    arvo      Added pre-order option (D_PreOrder).         *
*       09/02/99    arvo      Rewrote Directory class.                     *
*       08/17/99    arvo      Major restructuring.                         *
*       06/06/98    arvo      Added Directory class.                       *
*                                                                          *
*--------------------------------------------------------------------------*
* Copyright (C) 1999, James Arvo                                           *
*                                                                          *
* This program is free software; you can redistribute it and/or modify it  *
* under the terms of the GNU General Public License as published by the    *
* Free Software Foundation.  See http://www.fsf.org/copyleft/gpl.html      *
*                                                                          *
* This program is distributed in the hope that it will be useful, but      *
* WITHOUT EXPRESS OR IMPLIED WARRANTY of merchantability or fitness for    *
* any particular purpose.  See the GNU General Public License for more     *
* details.                                                                 *
*                                                                          *
***************************************************************************/
#ifndef __FILE_INCLUDED__
#define __FILE_INCLUDED__

#include <iostream>

using namespace std;

// "Or" the following flags together to indicate which objects are to
// be reported while traversing the directory structure.  For instance,
// if the flag is set to D_Dir | D_Link, then only directories and 
// links are reported, all other objects being skipped.
static const long
    D_Files    = 0x001, // Report normal files.
    D_Dirs     = 0x002, // Report directories.
    D_Links    = 0x004, // Report links.
    D_NonRec   = 0x008, // Do not follow subdirectories.
    D_Sort     = 0x010, // List objects alphabetically (not implemented).
    D_NoDots   = 0x020, // Omit files that begin with a dot.
    D_PreOrder = 0x040, // List files first, then follow sub-directories.
    D_All      = ( D_Files | D_Dirs | D_Links );

class Directory {

    public:
        Directory( const char *root = NULL, long = D_All );
        void Open( const char *root,        long = D_All );
        const char *ObjName ( char *buff = NULL ) const; 
        const char *FullPath( char *buff = NULL ) const; 
        int is_File() const;          // True iff current object is a file.
        int is_Dir () const;          // True iff current object is a directory.
        int is_Link() const;          // True iff current object is a link.
        int Level() const;            // The tree depth of this object.
        int Next();                   // Fetch the next object, if there is one.
        int operator==( const char *name   ) const; // Compare file name (not full path).
        int operator!=( const char *name   ) const; // Compare file name (not full path).
        int hasSuffix ( const char *suffix ) const;
        int End() const;  // True iff Next() will return 0.
       ~Directory();

    private:
        void Cleanup();
        void Push( const char *path );
        int  Pop ();
        int  Skip() const;

    private:
        void *state; // State is hidden so that fewer include
                     // files are needed with this one.

    friend ostream &operator<<( ostream &out, const Directory & );
    };

extern int is_File  ( const char *path );
extern int is_Dir   ( const char *path );
extern int is_Link  ( const char *path );
extern int can_Read ( const char *path );
extern int can_Write( const char *path );
extern int can_Exec ( const char *path );

extern ostream &operator<<( ostream &out, const Directory & );

#endif






