/***************************************************************************
* File.C                                                                   *
*                                                                          *
* Tools for traversing directory structures.                               *
*                                                                          *
*   HISTORY                                                                *
*       09/22/99    arvo      Added pre-order option (D_PreOrder).         *
*       09/02/99    arvo      Rewrote Directory class.                     *
*       06/06/98    arvo      Added Directory class.                       *
*                                                                          *
*--------------------------------------------------------------------------*
* Copyright (C) 1999, James Arvo                                           *
*                                                                          *
* This program is free software; you can redistribute it and/or modify it  *
* under the terms of the GNU General Public License as published by the    *
* Free Software Foundation.  See http://www.fsf.org/copyleft/gpl.html      *
*                                                                          *
* This program is distributed in the hope that it will be useful, but      *
* WITHOUT EXPRESS OR IMPLIED WARRANTY of merchantability or fitness for    *
* any particular purpose.  See the GNU General Public License for more     *
* details.                                                                 *
*                                                                          *
***************************************************************************/
#include <cstdio>
#include <cassert>
#include <istream>
#include <ostream>
#include <fstream>
#include <cstring>
#include <dirent.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <util/Char.h>
#include <util/File.h>

using namespace std;

// This class contains all the actual data used in traversing a
// directory structure.  It is only visible to the methods of the
// Directory class.
class State {
    public:
        State();
       ~State() {}
        State( const State *x ) { *this = *x; }
        void Init( const char *path = "", long flags = D_All );
        const char *Name() const;
        int Recursive() const { return !(flags & D_NonRec); }
    public:
        dirent *entry;
        DIR    *dir;
        State  *back;
        int     level;
        int     newdir;
        int     root;
        int     list_dir;
        int     list_file;
        long    flags;
        char    path[ 256 ];
    };

State::State()
    {
    Init();
    }

void State::Init( const char *the_path, long the_flags )
    {
    level     = 0;
    newdir    = 0;
    root      = 0;
    list_dir  = 1;
    list_file = 1;
    dir       = NULL;
    back      = NULL;
    flags     = the_flags;
    strcpy( path, the_path );
    }

const char *State::Name() const
    {
    if( dir == NULL || entry == NULL ) return "";
    return entry->d_name;
    }

Directory::Directory( const char *path, long flags )
    {
    state = new State;
    if( path != NULL ) Open( path, flags );
    }

void Directory::Open( const char *path, long flags )
    {
    Cleanup();
    State *s = (State *)state;
    s->Init( path, flags );
    s->root   = 1;
    s->newdir = 1;
    }

Directory::~Directory()
    {
    Cleanup();
    State *d = (State *)state;
    delete d;
    }

void Directory::Cleanup()
    {
    while( Pop() ); // Destroy any previous state.
    }

int Directory::operator==( const char *name ) const 
    {
    // Compare file name (not full path).
    return Match( ObjName(), name );
    }

int Directory::operator!=( const char *name ) const 
    {
    // Compare file name (not full path).
    return !Match( ObjName(), name );
    }

int Directory::hasSuffix( const char *suffix ) const
    {
    return ::hasSuffix( ObjName(), suffix );
    }

const char *Directory::ObjName( char *user_buff ) const
    {
    static char local_buff[128];
    char *buff =  ( user_buff == NULL ) ? local_buff : user_buff;
    State *s = (State *)state;
    strcpy( buff, s->Name() );
    return buff;
    }

const char *Directory::FullPath( char *user_buff ) const
    {
    static char local_buff[256];
    char *buff =  ( user_buff == NULL ) ? local_buff : user_buff;
    State *s = (State *)state;
    const char *object = s->Name();
    if( object == NULL || object[0] == NullChar ) 
         sprintf( buff, "%s"   , s->path         );
    else sprintf( buff, "%s/%s", s->path, object );
    return buff;
    }

int Directory::Level() const
    {
    State *s = (State *)state;
    return s->level;
    }

int Directory::is_File() const
    {
    return ::is_File( FullPath() );
    }

int Directory::is_Dir() const
    {
    return ::is_Dir( FullPath() );
    }

int Directory::is_Link() const
    {
    return ::is_Link( FullPath() );
    }

int Directory::Skip() const
    {
    State *s = (State *)state;
    const char *name = s->Name();
    if( name[0] == '.' && (s->flags & D_NoDots) ) return 1;
    if( Match( name, ".", ".."  ) ) return 1;
    return 0;
    }

// Create a new Directory node and copy the current directory
// information into it (including open parent directories).  When
// the current directory is exhausted, return to the one that 
// was pushed.
void Directory::Push( const char *path )
    {
    State *s = (State *)state;
    if( path != NULL && path[0] != NullChar )
        {
        State *old = new State( *s );
        s->back = old;
        s->dir  = opendir( path );
        s->root = 0;
        s->level++;
        strcpy( s->path, path );
        }
    }

// Return to a previously pushed directory, if there is one.
// Return 1 if so, and 0 if no more directories exist.
int Directory::Pop()
    {
    State *s = (State *)state;
    if( s->dir != NULL ) 
        {
        closedir( s->dir );  
        s->dir = NULL;
        }
    if( s->back != NULL )
        {
        State *old = s->back;
        *s = *old;
        delete old;
        return 1;
        }
    return 0;
    }

int Directory::Next()
    {
    State *s = (State *)state;

    // Descend into a "new" (i.e. unprocessed) directory if it 
    // is the root directory, or if recursion is enabled.
    // If the objects are to be listed in "pre-order", ensure that
    // files are listed first, then descend into sub-directories.
    if( s->newdir )
        {
        s->newdir = 0;
        if( s->root || s->Recursive() ) Push( FullPath() );
        if( s->flags & D_PreOrder )
            {
            s->list_file = 1;
            s->list_dir  = 0;
            }
        }

    // Search until we find the "next" object that is to be
    // reported.  Stop and return failure (i.e. 0) when the
    // root directory reaches the top of the stack.
    while( !s->root )
        {
        // Grab the next object from the current directory.
        // NULL indicates that we've depleted the list.
        assert( s->dir != NULL );
        s->entry = readdir( s->dir );
        if( s->entry == NULL )
            {
            if( !s->list_dir )
                {
                // Pre-order is on, and we have finished listing the files.
                // Re-open the directory and start the second pass.
                s->list_file = 0;
                s->list_dir  = 1;
                s->dir = opendir( s->path );
                }
            else
                {
                // No more files in this directory.  Move up a level.
                Pop();
                }
            }
        else if( Skip() )
            {
            // Keep searching;
            }
        else if( is_File() && s->list_file )
            {
            if( s->flags & D_Files ) return 1;
            }
        else if( is_Dir() && s->list_dir )
            {
            // If we must return now, mark the directory so
            // that it will be pushed as a result of the next
            // call to "Next".
            if( s->flags & D_Dirs ) 
                {
                s->newdir = 1;
                return 1;
                }
            else if( s->Recursive() ) Push( FullPath() );
            }
        else if( is_Link() )
            {
            if( s->flags & D_Links ) return 1;
            }
        }
    return 0;
    }

int Directory::End() const  // True iff Next() will return 0.
    {
    State *s = (State *)state;
    // There are no more objects to return if we are back to the root
    // directory after processing it; that is, it is no longer a "new"
    // directory.
    return s->root && !s->newdir;
    }

int is_File( const char *path )
    {
    struct stat buf;
    if( lstat( path, &buf ) != 0 ) return 0;
    return (buf.st_mode & S_IFMT) == S_IFREG;
    }

int is_Dir( const char *path )
    {
    struct stat buf;
    if( lstat( path, &buf ) != 0 ) return 0;
    return (buf.st_mode & S_IFMT) == S_IFDIR;
    }

int is_Link( const char *path )
    {
    struct stat buf;
    if( lstat( path, &buf ) != 0 ) return 0;
    return (buf.st_mode & S_IFMT) == S_IFLNK;
    }

int can_Exec( const char *path )
    {
    struct stat buf;
    if( lstat( path, &buf ) != 0 ) return 0;
    return buf.st_mode & S_IEXEC;
    }

int can_Read( const char *path )
    {
    struct stat buf;
    if( lstat( path, &buf ) != 0 ) return 0;
    return buf.st_mode & S_IREAD;
    }

int can_Write( const char *path )
    {
    struct stat buf;
    if( lstat( path, &buf ) != 0 ) return 0;
    return buf.st_mode & S_IWRITE;
    }


extern ostream &operator<<( ostream &out, const State &state )
    {
    out << form( "STATE %lX:\n", &state );
    out << form( "  back   = %lX\n", state.back   );
    out << form( "  level  = %d\n" , state.level  );
    out << form( "  newdir = %d\n" , state.newdir );
    out << form( "  root   = %d\n" , state.root   );
    out << form( "  flags  = %d\n" , state.flags  );
    out << form( "  path   = %s\n" , state.path   );
    out << endl;
    return out;
    }

ostream &operator<<( ostream &out, const Directory &dir )
    {
    State *s = (State *)dir.state;
    out << *s;
    return out;
    }



