/***************************************************************************
* TCL_Thread.C                                                             *
*                                                                          *
* The TCL_Thread class is the base class for all objects that wish to be   *
* awakened periodically, via a "WakeUp" call.  Each TCL_Thread object can  *
* set its own wakeup period and timeslice.  Relinquishing control is on    *
* a purely voluntary basis (consequently, these are not true threads).     *
*                                                                          *
*   Changes:                                                               *
*                                                                          *
*       08/06/2000   arvo    Rewritten for TCL wrapper.                    *
*       04/20/1998   arvo    Initial coding.                               *
*                                                                          *
***************************************************************************/

#include <cstdio>
#include <cstring>
#include <istream>
#include <ostream>
#include <fstream>
#include <util/Math.h>
#include <util/Timer.h>
#include <util/TCL_Thread.h>

Tcl_Interp *TCL_Thread::interp   = NULL;  // The Tcl interpreter.
char        TCL_Thread::cbid[64] =   "";  // ID of pending callback.

TCL_Thread::TCL_Thread()
    {
    active        = 0;
    alarm_time    = 0;
    time_slice    = MinTimeSlice;
    wakeup_period = MinDelay + time_slice;
    min_delay     = wakeup_period / 2;
    start_time    = Now();
    }

TCL_Thread::~TCL_Thread()
    {
    if( Running() ) StopThread();
    }

int TCL_Thread::Running() const
    {
    return active;
    }

void TCL_Thread::SetWakeupPeriod( int period ) // milliseconds
    {
    wakeup_period = MAX( period, MinDelay );
    }

void TCL_Thread::SetTimeSlice( int slice ) // milliseconds
    {
    time_slice = MIN( slice, MaxTimeSlice );
    }

void TCL_Thread::StartThread()
    {
    if( !Running() )
	{
        active = 1;
        Start();
        // Start the thread promptly (after a very short delay), then begin
        // periodic wakeups using the actual timings.
        WakeUpAfter( MinDelay );
	}
    }

void TCL_Thread::StopThread()
    {
    if( Running() )
	{
        // Cancel any pending callback.
        if( strlen( cbid ) > 0 ) 
	    {
            Tcl_VarEval( interp, "after cancel ", cbid, NULL );
            strcpy( cbid, "" );
	    }
        active = 0;
        Stop();
	}
    }

void TCL_Thread::WakeUpAfter( int delay )
    {
    static char buff[100];
    if( active && interp != NULL )
        {
        delay = ( delay == 0 ) ? UpdateTime : MAX( delay, MinDelay );
        min_delay  = delay / 2;
        alarm_time = Now();
        sprintf( buff, "%d { WakeUpThread %lu }", delay, this );
        Tcl_VarEval( interp, "after ", buff, NULL );
        // Save the cbid associated with the pending callback.  This can be
        // used to cancel the callback.
	strcpy( cbid, interp->result );
        }
    }

void TCL_Thread::WakeUpWrapper()
    {
    long now = Now();

    if( active )
        {
        // See if the callback really waited as long as we asked for.  It
        // appears that Tk makes a callback both immediately AND after the
        // requested time!  This causes the system to hang when there are
        // many active glyphs.
        long actual_delay = now - start_time;
        if( actual_delay < min_delay ) 
            {
            // printf( "actual = %d, min_delay = %d\n", actual_delay, min_delay );
            return;
            }
        }

    // Wake up the thread after recording the time.  This version of "WakeUp" is
    // a virtual method that is specific to this instance.  So, this is where the
    // actual work gets done.

    Lock();
      start_time = now;
      WakeUp();
    Unlock();

    // Now reset the timer to be the desired delay minus what we have
    // just used up.  If this becomes negative, use the minimum allowable 
    // delay, which is MinDelay.  This will ensure that Tk has 
    // time to process input.

    alarm_time = Now();
    int used   = alarm_time - start_time;
    int delay  = MAX( wakeup_period - used, MinDelay );

    // Reset the wakeup timer if this thread is "active", unless
    // periodic wakeups have been globally disabled.

    if( active ) WakeUpAfter( delay );
    }


int TCL_Thread::TimedOut() const // Nonzero if time slice has been used up.
    {
    return ( Now() - start_time ) > time_slice;
    }




// Static methods...

long TCL_Thread::Now()
    {
    // This method encapsulates the single timer used by the entire system.
    static Timer timer;
    return (long)timer.RealMilliseconds();
    }

void TCL_Thread::SetInterp( Tcl_Interp *interp_ )
    {
    // The TCL_Thread class needs access to the tcl interpreter, but does not
    // start one up itself.  Consequently, this procedure is used to tell TCL_Thread
    // about an existing interpreter.
    interp = interp_;
    Tcl_CreateCommand( interp, "WakeUpThread", TCL_Thread::WakeUpThread, NULL, NULL );
    }

int TCL_Thread::WakeUpThread( ClientData, Tcl_Interp *, int argc, const char **argv )
    {
    // This method is called directly from tcl, by means of an "after" command, to
    // wake up the given thread, giving it another time slice.  WakeupThread is
    // called with one argument, which is the actual address of the thread to wake
    // up -- i.e. the address of the TCL_Thread instance.
    const char *name = "WakeUpThread";
    if( argc != 2 )
        {
        sprintf( interp->result, "%s: Wrong number of arguments (%d)", name, argc );
        return TCL_ERROR;
	}
    // The wakeup callback has the address of the Thread to be awakened.
    TCL_Thread *thread = (TCL_Thread *)atol( argv[1] );
    thread->WakeUpWrapper();
    return TCL_OK;
    }



