/* *************************************************************************
* Copyright (c) 1999, Steve Smithies                                       *
*                                                                          *
* This program is free software; you can redistribute it and/or modify it  *
* under the terms of the GNU General Public License as published by the    *
* Free Software Foundation.  See http://www.fsf.org/copyleft/gpl.html      *
*                                                                          *
* This program is distributed in the hope that it will be useful, but      *
* WITHOUT EXPRESS OR IMPLIED WARRANTY of merchantability or fitness for    *
* any particular purpose.  See the GNU General Public License for more     *
* details.                                                                 *
*                                                                          *
* Revision History:                                                        *
*   v 1.0.0 Original Version, Steve Smithies 1999                          *
*   v 1.0.1 R. Zanibbi, March 15, 2002                                     *
*     - CharRec() altered to use new Font_File_Base_Directory() fn.        *
***************************************************************************/

#include <assert.h>
#include <string.h>
#include <strings.h>
#include <stdlib.h>
#include <stdio.h>

#include <tcl.h>
#include <tk.h>

#include "charrec.h"

// R. Zanibbi, April 29, 2004
// Function to return the number of available classes.
int CharRec::NumberClasses()
{
	return Recogniser -> NumSymbols();
}

// R. Zanibbi, April 29, 2004
// And a function to return the names of the available classes.
char* CharRec::GetSymbolNames()
{
	return Recogniser -> GetSymbolNames();
}

// we're now going to print out all the stuff we figured out, and
// let the calling program deal with it as it sees fit
// this kinda parallels the old Update_Rec_Data function
void CharRec::SaveStuffFiguredOut()
{
  ThisConf = 1.0;
  StrokesCountedInRecognisedGlyphs = 0;

  Get_Seq(&StrokeCombos, CurrentSeq, BestSeqNum);
  StuffFiguredOut[0] = 0;

  sprintf(StuffFiguredOut,"{");

  int i=0;
  while (CurrentSeq[i] != -1)
    {
      sprintf(StuffFiguredOut,"%s%d ",StuffFiguredOut, CurrentSeq[i++]);
    }

  sprintf(StuffFiguredOut,"%s} {",StuffFiguredOut);

  // we're only interested in the first character recognised
  for (int i=0;i<StrokeFromCollections[0].Length();i++)
    sprintf(StuffFiguredOut,"%s%d ",StuffFiguredOut, StrokeFromCollections[0].Item(i));

  sprintf(StuffFiguredOut,"%s} {%s}\n",
	  StuffFiguredOut, 
	  PreCalc[0][CurrentSeq[0]].OtherChoices);
}


// Take a series of strokes, gluing them together into a Multiline
void CharRec::Glue_Together_Strokes(int StartStrokeNum, 
				    int NumberOfStrokes,
				    const Multiline *StrokeM,
				    Multiline &ML)
{
  int StrokeNum;
  Vec2        P;  // A 2d point
  StrokeNum = StartStrokeNum;

  ML.Clear();

  for (int j=0;j<NumberOfStrokes;j++) { // glue together j strokes for this character	
    ML << StrokeM[StrokeNum++]; // "<<" doesn't add a jump between.  += does
    if (j<(NumberOfStrokes-1)) {
      ML << Multiline::Jump; // add a jump if we're not at the end
    }
  }
}

bool CharRec::MultilinesOverlap(Multiline &M1, Multiline &M2)
{
  bool WeOverLap = false;

  if ((M1.Box()*M2.Box()).Area() == 0) // no intersection of bounding boxes, don't waste time doing
    return false;		     // slower overlap testing
  
  for (int icompnum=1;icompnum<M1.Points();icompnum++) {
    for (int jcompnum=1;jcompnum<M2.Points();jcompnum++) {
      Position_2D A,B,C,D;
      A.x = M1.Vertex(icompnum-1).X();
      A.y = M1.Vertex(icompnum-1).Y();
      B.x = M1.Vertex(icompnum).X();
      B.y = M1.Vertex(icompnum).Y();

      C.x = M2.Vertex(jcompnum-1).X();
      C.y = M2.Vertex(jcompnum-1).Y();
      D.x = M2.Vertex(jcompnum).X();
      D.y = M2.Vertex(jcompnum).Y();

      if (LineIntersectTest(A,B,C,D)) {
	WeOverLap = true;
      }
    } // for jcompnum
  } // icompnum


  return WeOverLap;
}

void CharRec::Allocate_Do_Recognise_Memory()
{
  StrokeLength = Calloc_int(MAX_STROKES_TO_PROCESS_AT_ONCE);
  StrokeMultiline = new Multiline[MAX_STROKES_TO_PROCESS_AT_ONCE];
}

void CharRec::Tidy_Up_Memory()
{
  Free_2D_char(ComboStrings, Combo_Size(&StrokeCombos));
  Free_int(StrokeLength);
  delete[] StrokeMultiline;
  for (int i=0;i<=StrokeCount;i++) {
    for (int j=0;j<=StrokeCount;j++) {
      if (PreCalc[i][j].WhatChar != NULL) {
	free(PreCalc[i][j].WhatChar);
      }
    }
    free(PreCalc[i]);
  }
  free(PreCalc);
  free(CurrentSeq);
}

void CharRec::Find_Best_Combo()
{
  // //////////////////////////////////////////////////////////////////////////////////////////////////
  // Now find the "best" combo
    
  BestSeqConf = 0;
  BestSeqNum = -1;
  
  for (int i=0;i<Combo_Size(&StrokeCombos);i++) { // for each sequence
    ThisConf = 1.0;
    StrokesCountedInRecognisedGlyphs = 0;

    Get_Seq(&StrokeCombos, CurrentSeq, i);
    StrokesUsed = 0;
    SeqPos = 0;
    TotalStrokeCount = 0;
    while (CurrentSeq[SeqPos] != -1) {

      STRSPACECAT(ComboStrings[i], PreCalc[StrokesUsed][CurrentSeq[SeqPos]].WhatChar);

      StrokesCountedInRecognisedGlyphs +=  PreCalc[StrokesUsed][CurrentSeq[SeqPos]].NumberOfStrokes;
       
      ThisConf = MIN(PreCalc[StrokesUsed][CurrentSeq[SeqPos]].Confidence, ThisConf); // \cite{turban92:exper_system_applied_artif_intel}

      for (int j=0;j<CurrentSeq[SeqPos];j++) {
	TotalStrokeCount += StrokeMultiline[StrokesUsed+j].Comps();
      }


      StrokesUsed += CurrentSeq[SeqPos];
      SeqPos++;
    } // end of while
    
    if ((BestSeqNum == -1)||(ThisConf > BestSeqConf)) {
      BestSeqNum = i;
      BestSeqConf = ThisConf;
    }


  } // for i...numberofcombos
}

void CharRec::Do_Precalculate()
{
  int i,j;

  // //////////////////////////////////////////////////////////////////////////////////////////////////
  // Cunning precalculate bit.
  PreCalc = (PreCalcData **)calloc(StrokeCount+1, sizeof(PreCalcData *));
  if (PreCalc == NULL) {
    fprintf(stderr,"recognise.cc:Do_Recognise -  Error allocating PreCalc\n");
    exit(1);
  }

  for (i=0;i<=StrokeCount;i++) {
    PreCalc[i] = (PreCalcData *)calloc(1+StrokeCount, sizeof(PreCalcData));
    if (PreCalc[i] == NULL) {
      fprintf(stderr,"recognise.cc:Do_Recognise -  Error allocating PreCalc[%d]\n",i);
      exit(1);
    }
    for (j=0;j<=StrokeCount;j++) {
      PreCalc[i][j].WhatChar = NULL;
    }
  }

  for (i=0;i<StrokeCount;i++) {
    for (j=1;j<=(StrokeCount - i);j++) {
      int FoundOne;

      Glue_Together_Strokes(i, // start
			    j, // number
			    StrokeMultiline, // stroke info
			    PreCalc[i][j].Stroke // multiline to put it into
			    );

      Recogniser->Recognize( PreCalc[i][j].Stroke, LetterProf );

      FoundOne = 0;
      for (int num=0;num<LetterProf.NumChoices();num++) {
	char *WI = strdup(LetterProf.GetSymbol(num)->Name());

	if ( (strcmp(WI,"{") == 0) || (strcmp(WI,"}") == 0) ||
	     (strcmp(WI,"[") == 0) || (strcmp(WI,"]") == 0) ) { // braces confuse Tcl/Tk :( WOMBLE!!! Fix somehow! (maybe)
	  // skip to next iteration of this loop
	  continue;
	}

	if ( (!FoundOne) ) {
	  PreCalc[i][j].Confidence = LetterProf.GetProb(num);
	  if (isnan(PreCalc[i][j].Confidence))
	    fprintf(stderr,"NaN for confidence of %s\n",LetterProf.GetSymbol(num)->Name());
	  PreCalc[i][j].WhatChar = strdup(LetterProf.GetSymbol(num)->Name());
	  PreCalc[i][j].NumberOfStrokes = LetterProf.GetSymbol(num)->Stroke().Comps();
	  PreCalc[i][j].BoundingBox = PreCalc[i][j].Stroke.Box();
	  FoundOne = 1;
	}
	
	STRSPACECAT(PreCalc[i][j].OtherChoices, WI);
	
      } // for num...

      if (!FoundOne) {
	PreCalc[i][j].NumberOfStrokes =  LetterProf.GetSymbol(0)->Stroke().Comps();
	PreCalc[i][j].Confidence = 0.0; // force low confidence
	PreCalc[i][j].WhatChar = strdup(LetterProf.GetSymbol(0)->Name());
	PreCalc[i][j].BoundingBox = PreCalc[i][j].Stroke.Box();
      }

    }
  }
}

void CharRec::Generate_Stroke_Combos()
{
  // //////////////////////////////////////////////////////////////////////////////////////////////////
  // Generate stroke combos
  Generate(StrokeCount, &StrokeCombos, MaxStrokeCount);

  CurrentSeq = (ComboSequence *)calloc(StrokeCount+1,sizeof(int));

  ComboStrings = Calloc_2D_char(Combo_Size(&StrokeCombos), MAX_STRING_LENGTH*StrokeCount);
}

void CharRec::Determine_Overlaps()
{
  // //////////////////////////////////////////////////////////////////////////////////////////////////
  // Determine which strokes overlap which other strokes, and combine them
  // Look for intersections between strokes.

  int ChangeMade;
  int WeOverlap;

  do {
    ChangeMade = 0;
    for (int i=0;((!ChangeMade)&&(i<(StrokeCount-1)));i++) {
      for (int j=i+1;((!ChangeMade)&&(j<StrokeCount));j++) {
	// Does stroke #i overlap #j?
	// Do this by running along components in both i and j

	WeOverlap = 0;
	WeOverlap = MultilinesOverlap(StrokeMultiline[i],StrokeMultiline[j]);

	if (WeOverlap) {
	  StrokeMultiline[i] += StrokeMultiline[j]; // add j  to the end of i, with a jump between (+=)
	  // add in combos from added stroke
	  for (int CollCount = 0; CollCount < StrokeFromCollections[j].Length() ; CollCount++) { // badger
	    StrokeFromCollections[i].AddItem( StrokeFromCollections[j].Item(CollCount) );
	  }

	  // shuffle everything down one to fill gap
	  for (int k=j;k<(StrokeCount-1);k++) {
	    StrokeMultiline[k] = StrokeMultiline[k+1];
	    StrokeFromCollections[k].Clear();
	    for (int CollCount = 0; CollCount < StrokeFromCollections[k+1].Length() ; CollCount++) { // badger
	      StrokeFromCollections[k].AddItem( StrokeFromCollections[k+1].Item(CollCount) );
	    }
	    StrokeFromCollections[k+1].Clear();
	  }
	  StrokeCount--;
	  ChangeMade = 1;
	} else {
	  // do nothing
	}
      } // for j
    } // for i
  } while (ChangeMade); // repeat until no more changes are made
}

void CharRec::Display_All_Strokes()
{
  // //////////////////////////////////////////////////////////////////////////////////////////////////
  // Decorate the screen
  int i,j;

  fprintf(stderr,"all %d strokes:\n",StrokeCount);
  for (i=0;i<StrokeCount;i++) {
    fprintf(stderr,"#%d ",i);
    for (j=0;j<StrokeMultiline[i].Points();j++) {
      fprintf(stderr,"%d,%d ",
	      (int)StrokeMultiline[i].Vertex(j).X(),
	      (int)StrokeMultiline[i].Vertex(j).Y());
    }
    fprintf(stderr,"\n");
    for (j=0;j<StrokeFromCollections[i].Length();j++) {
      fprintf(stderr,"%d ",StrokeFromCollections[i].Item(j));
    }
    fprintf(stderr,"\n");
  }
}

void CharRec::Initialise_Strokes()
{
  // Empty out strokes thing
  StrokeCount = 0;
  StrokeLength[StrokeCount] = 0;
  StrokeMultiline[StrokeCount].Clear();
  //  StrokeFromCollection[StrokeCount] = -1; badger
  StrokeFromCollections[StrokeCount].Clear();
}

void CharRec::GoDoYourFunkyThing()
{
  Generate_Stroke_Combos();
  Do_Precalculate();
  Find_Best_Combo();
  SaveStuffFiguredOut();
  Tidy_Up_Memory();
}

CharRec::CharRec() 
{
  
  fprintf(stderr,"Initialising recogniser, please wait...\n");

  // R. Zanibbi March 15, 2002: using the FFES base dir returned
  // by Font_File_Base_Directory.
  //
  // R. Zanibbi April 24, 2003: updating to use revised CIT4.3.1
  char dataDir[256] = "";
  strcat(dataDir,Font_File_Base_Directory());
  strcat(dataDir,"/CIT4.3.1/data");
  MaxStrokeCount = Max_Strokes_In_Font(dataDir);
  Recogniser = new FeatureRecognizer(dataDir);
  fprintf(stderr,"Recogniser ready.\n");
}

CharRec::~CharRec()
{
  delete Recogniser;
}

/*
void CharRec::DoSpecificCollection(
				    )
{
      Allocate_Do_Recognise_Memory();
      Initialise_Strokes();


      int OriginalCollection;
      scanf("%d",&OriginalCollection);
      StrokeFromCollections[StrokeCount].AddItem(OriginalCollection);

      int PointPairs;
      scanf("%d",&PointPairs);
      Multiline ML;

      for (int j=0;j<PointPairs;j++)
	{
	  int x,y;
	  scanf("%d %d",&x,&y);
	  if ((x==-1)||(y==-1))
	    {
	      ML << Multiline::Jump;
	      StrokeLength[StrokeCount]++;
	    }
	  else 
	    {
	      Vec2 P (x,y);
	      ML << P;
	    }
	} // pointpairs
      StrokeMultiline[StrokeCount] += ML;
      StrokeCount++;

      GoDoYourFunkyThing();
}
*/      

void CharRec::DoNormalRec(int NumberOfMultilines,
			  int *OriginalCollection,
			  int *PointCounts,
			  int *Points)
{
  Allocate_Do_Recognise_Memory();
  Initialise_Strokes();
  
  int tp = 0;
  for (int i=0;i<NumberOfMultilines;i++)
    tp += PointCounts[i];

  int pc = 0;
  for (int i=0;i<NumberOfMultilines;i++)
    {
      StrokeMultiline[StrokeCount].Clear();
      StrokeFromCollections[StrokeCount].Clear();      

      StrokeFromCollections[StrokeCount].AddItem(OriginalCollection[i]);

      Multiline ML;
      int PointPairs = PointCounts[i];

      for (int j=0;j<PointPairs;j++)
	{
	  int x,y;

	  x = Points[ pc++ ];
	  y = Points[ pc++ ];
	  if ((x==-1)||(y==-1))
	    {
	      ML << Multiline::Jump;
	      StrokeLength[StrokeCount]++;
	    }
	  else 
	    {
	      Vec2 P (x,y);
	      ML << P;
	    }
	} // for j=0...pointpairs
      StrokeMultiline[StrokeCount] += ML;
      StrokeCount++;
    } // for i=0;i<number of multilines

  Determine_Overlaps();
  GoDoYourFunkyThing();
}

void CharRec::GetResultForTclTk(char *Dest)
{
  strcpy(Dest,StuffFiguredOut);
}
